@@ -31,14 +31,16 @@
 #
 
 package HTML::Mason::Request;
-
+$HTML::Mason::Request::VERSION = '1.58';
 use strict;
 use warnings;
 
 use File::Spec;
+use HTML::Mason::Cache::BaseCache;
 use HTML::Mason::Plugin::Context;
 use HTML::Mason::Tools qw(can_weaken read_file compress_path load_pkg pkg_loaded absolute_comp_path);
 use HTML::Mason::Utils;
+use Log::Any qw($log);
 use Class::Container;
 use base qw(Class::Container);
 
@@ -80,12 +82,12 @@ BEGIN
 
          data_cache_api =>
          { parse => 'string', default => '1.1', type => SCALAR,
-           regex => qr/^(?:1\.0|1\.1)$/,
-           descr => "Data cache API to use: 1.0 or 1.1" },
+           regex => qr/^(?:1\.0|1\.1|chi)$/,
+           descr => "Data cache API to use: 1.0, 1.1, or chi" },
 
          data_cache_defaults =>
          { parse => 'hash_list', type => HASHREF|UNDEF, optional => 1,
-           descr => "A hash of default parameters for Cache::Cache" },
+           descr => "A hash of default parameters for Cache::Cache or CHI" },
 
          declined_comps =>
          { type => HASHREF, optional => 1,
@@ -112,6 +114,10 @@ BEGIN
            regex => qr/^(?:output|fatal)$/,
            descr => "How error conditions are manifest (output or fatal)" },
 
+         component_error_handler =>
+         { parse => 'code', type => CODEREF|SCALAR, default => \&rethrow_exception,
+           descr => "A subroutine reference called on component compilation or runtime errors" },
+
          max_recurse =>
          { parse => 'string', default => 32, type => SCALAR,
            descr => "The maximum recursion depth for component, inheritance, and request stack" },
@@ -145,6 +151,7 @@ BEGIN
 my @read_write_params;
 BEGIN { @read_write_params = qw(
                                 autoflush
+                                component_error_handler
                                 data_cache_api
                                 data_cache_defaults
                                 dhandler_name
@@ -219,7 +226,8 @@ my %plugin_loaded;
 sub _initialize {
     my ($self) = @_;
 
-    local $SIG{'__DIE__'} = \&rethrow_exception;
+    local $SIG{'__DIE__'} = $self->component_error_handler
+        if $self->component_error_handler;
 
     eval {
         # Check the static_source touch file, if it exists, before the
@@ -234,6 +242,8 @@ sub _initialize {
         if (!ref($request_comp)) {
             $request_comp =~ s{/+}{/}g;
             $self->{top_path} = $path = $request_comp;
+            $log->debugf("top path is '%s'", $self->{top_path})
+                if $log->is_debug;
 
             my $retry_count = 0;
             search: {
@@ -246,6 +256,8 @@ sub _initialize {
                     if ( $request_comp = $self->interp->find_comp_upwards($path, $self->dhandler_name) ) {
                         my $parent_path = $request_comp->dir_path;
                         ($self->{dhandler_arg} = $self->{top_path}) =~ s{^$parent_path/?}{};
+                        $log->debugf("found dhandler '%s', dhandler_arg '%s'", $parent_path, $self->{dhandler_arg})
+                            if $log->is_debug;
                     }
                 }
 
@@ -299,7 +311,7 @@ sub _initialize {
                     #
                     {
                         no strict 'refs';
-                        unless (defined(%{$plugin . "::"})) {
+                        unless (keys %{$plugin . "::"}) {
                             eval "use $plugin;";
                             die $@ if $@;
                         }
@@ -331,7 +343,7 @@ sub _initialize {
 sub use_dhandlers
 {
     my $self = shift;
-    return defined $self->{dhandler_name} and length $self->{dhandler_name};
+    return (defined $self->{dhandler_name} and length $self->{dhandler_name});
 }
 
 sub alter_superclass
@@ -344,7 +356,8 @@ sub alter_superclass
     my $isa_ref;
     {
         no strict 'refs';
-        $isa_ref = \@{"$class\::ISA"};
+        my @isa = @{ $class . '::ISA' };
+        $isa_ref = \@isa;
     }
 
     # handles multiple inheritance properly and preserve
@@ -358,13 +371,18 @@ sub alter_superclass
             if ( $old_super ne $new_super )
             {
                 $isa_ref->[$x] = $new_super;
-
-                $class->valid_params( %{ $class->valid_params } );
             }
 
             last;
         }
     }
+
+    {
+        no strict 'refs';
+        @{ $class . '::ISA' } = @{ $isa_ref };
+    }
+
+    $class->valid_params( %{ $class->valid_params } );
 }
 
 sub exec {
@@ -374,8 +392,8 @@ sub exec {
     # at the bottom of _initialize(); just return.
     return unless $self->initialized();
 
-    # All errors returned from this routine will be in exception form.
-    local $SIG{'__DIE__'} = \&rethrow_exception;
+    local $SIG{'__DIE__'} = $self->component_error_handler
+        if $self->component_error_handler;
 
     # Cheap way to prevent users from executing the same request twice.
     #
@@ -409,6 +427,9 @@ sub exec {
     $self->{request_buffer} = $self->interp->preallocated_output_buffer;
     $self->{request_buffer} = '';
 
+    $log->debugf("starting request for '%s'", $self->request_comp->title)
+        if $log->is_debug;
+
     eval {
         # Build wrapper chain and index.
         my $request_comp = $self->request_comp;
@@ -437,7 +458,7 @@ sub exec {
             tie *SELECTED, 'Tie::Handle::Mason';
 
             my $old = select SELECTED;
-            my $mods = {base_comp => $request_comp, store => \($self->{request_buffer})};
+            my $mods = {base_comp => $request_comp, store => \($self->{request_buffer}), flushable => 1};
 
             if ($self->{has_plugins}) {
                 my $context = bless
@@ -485,6 +506,9 @@ sub exec {
         }
     };
 
+    $log->debugf("finishing request for '%s'", $self->request_comp->title)
+        if $log->is_debug;
+
     # Purge code cache if necessary.
     $self->interp->purge_code_cache;
 
@@ -647,8 +671,24 @@ sub cache
     if ($self->data_cache_defaults) {
         %options = (%{$self->data_cache_defaults}, %options);
     }
-    $options{namespace}   ||= compress_path($self->current_comp->comp_id);
-    $options{cache_root}  ||= $self->interp->cache_dir;
+
+    # If using the CHI API, just create and return a CHI handle. Namespace will be escaped by CHI.
+    if ($self->data_cache_api eq 'chi') {
+        my $chi_root_class = delete($options{chi_root_class}) || 'CHI';
+        load_pkg($chi_root_class);
+        if (!exists($options{namespace})) {
+            $options{namespace} = $self->current_comp->comp_id;
+        }
+        if (!exists($options{driver}) && !exists($options{driver_class})) {
+            $options{driver} = $self->interp->cache_dir ? 'File' : 'Memory';
+            $options{global} = 1 if $options{driver} eq 'Memory';            
+        }
+        $options{root_dir} ||= $self->interp->cache_dir;
+        return $chi_root_class->new(%options);
+    }
+
+    $options{cache_root} ||= $self->interp->cache_dir;
+    $options{namespace}  ||= compress_path($self->current_comp->comp_id);
 
     # Determine cache_class, adding 'Cache::' in front of user's
     # specification if necessary.
@@ -1179,6 +1219,7 @@ sub print
 #
 sub comp {
     my $self = shift;
+    my $log_is_debug = $log->is_debug;
 
     # Get modifiers: optional hash reference passed in as first argument.
     # Merge multiple hash references to simplify user and internal usage.
@@ -1205,13 +1246,18 @@ sub comp {
     error "$depth levels deep in component stack (infinite recursive call?)\n"
         if $depth >= $self->{max_recurse};
 
+    # Log start of component call.
+    #
+    $log->debugf("entering component '%s' [depth %d]", $comp->title(), $depth)
+        if $log_is_debug;
+
     # Keep the same output buffer unless store modifier was passed. If we have
     # a filter, put the filter buffer on the stack instead of the regular buffer.
     #
     my $filter_buffer = '';
     my $top_buffer = defined($mods{store}) ? $mods{store} : $self->{top_stack}->[STACK_BUFFER];
     my $stack_buffer = $comp->{has_filter} ? \$filter_buffer : $top_buffer;
-    my $flushable = exists $mods{flushable} ? $mods{flushable} : 1;
+    my $flushable = exists $mods{flushable} ? $mods{flushable} : ($self->{top_stack}->[STACK_BUFFER_IS_FLUSHABLE] && ! defined($mods{store})) ;
 
     # Add new stack frame and point dynamically scoped $self->{top_stack} at it.
     push @{ $self->{stack} },
@@ -1290,6 +1336,11 @@ sub comp {
     # they don't get cleaned up until the component exits.
     pop @{ $self->{stack} };
 
+    # Log end of component call.
+    #
+    $log->debug(sprintf("exiting component '%s' [depth %d]", $comp->title(), $depth))
+        if $log_is_debug;
+
     # Repropagate error if one occurred, otherwise return result.
     rethrow_exception $error if $error;
     return $wantarray ? @result : $result[0];
@@ -1301,7 +1352,7 @@ sub comp {
 sub scomp {
     my $self = shift;
     my $buf;
-    $self->comp({store => \$buf, flushable => 0},@_);
+    $self->comp({store => \$buf},@_);
     return $buf;
 }
 
@@ -1401,15 +1452,6 @@ sub request_args
 *top_args = \&request_args;
 *top_comp = \&request_comp;
 
-# deprecated in 1.1x
-sub time
-{
-    my ($self) = @_;
-    my $time = $self->interp->current_time;
-    $time = time() if $time eq 'real';
-    return $time;
-}
-
 #
 # Subroutine called by every component while in debug mode, convenient
 # for breakpointing.
@@ -1499,8 +1541,14 @@ sub _compute_base_comp_for_frame {
     return $frame->[STACK_BASE_COMP];
 }
 
-package Tie::Handle::Mason;
+sub log
+{
+    my ($self) = @_;
+    return $self->current_comp->logger();
+}
 
+package Tie::Handle::Mason;
+$Tie::Handle::Mason::VERSION = '1.58';
 sub TIEHANDLE
 {
     my $class = shift;
@@ -1598,21 +1646,41 @@ as well.
 
 =item data_cache_api
 
-The C<$m-E<gt>cache> API to use. '1.1', the default, indicates the newer
-API L<documented in this manual|HTML::Mason::Request/item_cache>.
-'1.0' indicates the old API documented in 1.0x and earlier. This
-compatibility layer is provided as a convenience for users upgrading
-from older versions of Mason, but will not be supported indefinitely.
+The C<$m-E<gt>cache> API to use:
+
+=over
+
+=item *
+
+'1.1', the default, indicates a C<Cache::Cache> based API.
+
+=item *
+
+'chi' indicates a C<CHI> based API.
+
+=item *
+
+'1.0' indicates the custom cache API used in Mason 1.0x and
+earlier. This compatibility layer is provided as a convenience for
+users upgrading from older versions of Mason, but will not be
+supported indefinitely.
+
+=back
 
 =item data_cache_defaults
 
 A hash reference of default options to use for the C<$m-E<gt>cache>
-command. For example, to use the C<MemoryCache> implementation
-by default,
+command.  For example, to use Cache::Cache's C<MemoryCache>
+implementation by default:
 
     data_cache_defaults => {cache_class => 'MemoryCache'}
 
-These settings are overriden by options given to particular
+To use the CHI C<FastMmap> driver by default:
+
+    data_cache_api      => 'CHI',
+    data_cache_defaults => {driver => 'FastMmap'},
+
+These settings are overridden by options given to particular
 C<$m-E<gt>cache> calls.
 
 =item dhandler_name
@@ -1637,7 +1705,8 @@ I<text> - a multi-line text format
 
 =item *
 
-I<line> - a single-line text format, with different pieces of information separated by tabs (useful for log files)
+I<line> - a single-line text format, with different pieces of
+information separated by tabs (useful for log files)
 
 =item *
 
@@ -1667,6 +1736,19 @@ L<CGI|HTML::Mason::CGIHandler> is I<output>, causing the error to be
 displayed in the browser.  The default for standalone mode is
 I<fatal>.
 
+=item component_error_handler
+
+A code reference used to handle errors thrown during component
+compilation or runtime. By default, this is a subroutine that turns
+non-exception object errors in components into exceptions. If this
+parameter is set to a false value, these errors are simply rethrown
+as-is.
+
+Turning exceptions into objects can be expensive, since this will
+cause the generation of a stack trace for each error. If you are using
+strings or unblessed references as exceptions in your code, you may
+want to turn this off as a performance boost.
+
 =item max_recurse
 
 The maximum recursion depth for the component stack, for the request
@@ -1795,13 +1877,17 @@ example inside a plugin's C<start_request_hook()> method, where we
 have created a request but it does not yet know anything about the
 component being called.
 
-=item cache (cache_class=>'...', [cache_options])
+=item cache
 
 =for html <a name="item_cache"></a>
 
-C<$m-E<gt>cache> returns a new L<cache
-object|HTML::Mason::Cache::BaseCache> with a namespace specific to
-this component.
+C<$m-E<gt>cache> returns a new L<cache object|HTML::Mason::Cache::BaseCache> with a
+namespace specific to this component. The parameters to and return value from
+C<$m-E<gt>cache> differ depending on which L<data_cache_api> you are using.
+
+=over
+
+=item If data_cache_api = 1.1 (default)
 
 I<cache_class> specifies the class of cache object to create. It
 defaults to C<FileCache> in most cases, or C<MemoryCache> if the
@@ -1809,17 +1895,27 @@ interpreter has no data directory, and must be a backend subclass of
 C<Cache::Cache>. The prefix "Cache::" need not be included.  See the
 C<Cache::Cache> package for a full list of backend subclasses.
  
-I<cache_options> may include any valid options to the new() method of
-the cache class. e.g. for C<FileCache>, valid options include
-default_expires_in and cache_depth.
+Beyond that, I<cache_options> may include any valid options to the new() method of the
+cache class. e.g. for C<FileCache>, valid options include C<default_expires_in> and
+C<cache_depth>.
 
-See the L<data caching|HTML::Mason::Devel/data caching> section of the developer's manual for a caching tutorial and examples. See the
-L<HTML::Mason::Cache::BaseCache|HTML::Mason::Cache::BaseCache>
-documentation for a method reference.
+See L<HTML::Mason::Cache::BaseCache|HTML::Mason::Cache::BaseCache> for
+information about the object returned from C<$m-E<gt>cache>.
 
-Note: users upgrading from 1.0x and earlier can continue to use the
-old C<$m-E<gt>cache> API by setting L<data_cache_api|HTML::Mason::Params/data_cache_api> to '1.0'.  This
-support will be removed at a later date.
+=item If data_cache_api = CHI
+
+I<chi_root_class> specifies the factory class that will be called to
+create cache objects. The default is 'CHI'.
+
+I<driver> specifies the driver to use, for example C<Memory> or
+C<FastMmap>.  The default is C<File> in most cases, or C<Memory> if
+the interpreter has no data directory.
+
+Beyond that, I<cache_options> may include any valid options to the
+new() method of the driver. e.g. for the C<File> driver, valid options
+include C<expires_in> and C<depth>.
+
+=back
 
 =item cache_self ([expires_in => '...'], [key => '...'], [get_options], [cache_options])
 
@@ -2001,7 +2097,7 @@ Note that we ignore both the original and the final return value.
     }
     ...
 
-Here is a piece of code that traps all errors occuring anywhere in a
+Here is a piece of code that traps all errors occurring anywhere in a
 component or its children, e.g. for the purpose of handling
 application-specific exceptions. This is difficult to do with a manual
 C<eval> because it would have to span multiple code sections and the
@@ -2096,6 +2192,15 @@ because it will not try to evaluate the content.
 Returns the number of this request, which is unique for a given
 request and interpreter.
 
+=item current_args
+
+=for html <a name="item_current_args"></a>
+
+Returns the arguments passed to the current component. When called in
+scalar context, a hash reference is returned.  When called in list
+context, a list of arguments (which may be assigned to a hash) is
+returned.
+
 =item current_comp
 
 =for html <a name="item_current_comp"></a>
@@ -2228,6 +2333,14 @@ relative to the current component.
 
 See the L<subrequests|HTML::Mason::Devel/subrequests> section of the developer's manual for more information about subrequests.
 
+=item log
+
+=for html <a name="item_log"></a>
+
+Returns a C<Log::Any> logger with a log category specific to the
+current component.  The category for a component "/foo/bar" would be
+"HTML::Mason::Component::foo::bar".
+
 =item notes (key, value)
 
 =for html <a name="notes"></a>
@@ -2419,14 +2532,4 @@ exceptions, like this:
 
 =back
 
-=head1 AUTHORS
-
-Jonathan Swartz <swartz@pobox.com>, Dave Rolsky <autarch@urth.org>, Ken Williams <ken@mathforum.org>
-
-=head1 SEE ALSO
-
-L<HTML::Mason|HTML::Mason>,
-L<HTML::Mason::Devel|HTML::Mason::Devel>,
-L<HTML::Mason::Component|HTML::Mason::Component>
-
 =cut
